#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program  
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
# 
#-------------------------------------------------------------------------------------------------
from os import path, environ, tempnam, mkdir
import sys

class ConfigError(Exception):
    pass

class ConfigNotSetError(ConfigError):
    pass

class BadConfigKeyError(ConfigError):
    pass

DEFAULTS = dict(
    listen_address='0.0.0.0',
    listen_port=3001,
    monitor_password="XIVmonitor",
    manager_password="XIVmanager",
    require_authentication=True,
    log_file_name="xiv_mpio_agent.log",
    child_log_file_name="xiv_mpio_agent_c.log",
    log_level="info",
    snapshot_polling_interval=5.0,
    priority='below_normal'
)

PRIORITIES = dict(
    IDLE=1,
    BELOW_NORMAL=2,
    NORMAL=3,
    ABOVE_NORMAL=4,
    HIGH=5,
    REALTIME=6
)

class Config(object):
    SERVICE_NAME = 'XIVmpioAgent'
    SERVICE_DISPLAY_NAME = 'XIV MPIO Agent'
    SERVICE_DESCRIPTION = 'Monitors multipath devices and detects path problems'

    def __init__(self):
        self._registry = None
        self.data_dir = None
        self.log_dir = None
        self.priority = None

    def set(self, registry):
        self._registry = registry
        for key, default_value in DEFAULTS.items():
            setattr(self, key, self._registry.load_value(key, default_value))

        self._set_log_dir()
        self._set_data_dir()
        self._set_priority()

    def _set_log_dir(self):
        self.log_dir = self._registry.load_value('log_directory', None)
        if self.log_dir is None:
            if sys.platform == 'win32':
                self.log_dir = environ['TEMP']
            else:
                self.log_dir = path.dirname(tempnam())

    def _set_data_dir(self):
        self.data_dir = self._registry.load_value('data_directory', None)
        if self.data_dir is None:
            from hak.common.definitions import PROJECT_ROOT_PATH
            root_path = PROJECT_ROOT_PATH
            if root_path is None:
                raise ConfigError("failed to find project root")
            self.data_dir = path.realpath(path.join(root_path, "data"))
        if not path.exists(config.data_dir):
            mkdir(config.data_dir)

    def _set_priority(self):
        self.priority = self.priority.upper()
        if self.priority not in PRIORITIES:
            raise BadConfigKeyError('priority is set to unknown value (%s)' % self.priority)
        self.priority = PRIORITIES[self.priority]

    def __getattr__(self, attr):
        if self._registry is None:
            raise ConfigNotSetError
        if not hasattr(self, attr):
            raise BadConfigKeyError(attr)
        return self[attr]

    def set_value(self, attr, value):
        if self._registry is None:
            raise ConfigNotSetError
        if not hasattr(self, attr):
            raise BadConfigKeyError(attr)
        setattr(self, attr, value)
        self._registry.save_value(attr, value)

    def get_path_from_data_dir(self, *args):
        return path.realpath(path.join(self.data_dir, *args))

    def init_logging(self, log_file_name_key='log_file_name'):
        from twisted.python import log
        import logging

        log_file_name = getattr(self, log_file_name_key)
        if log_file_name != 'stdout':
            if not path.exists(self.log_dir):
                mkdir(self.log_dir)

            log_file_path = path.join(self.log_dir, log_file_name)

            from logging.handlers import RotatingFileHandler
            handler = RotatingFileHandler(log_file_path, maxBytes=1000000,
                                          backupCount=10)
        else:
            handler = logging.StreamHandler()

        formatter = self.get_logging_formatter()
        handler.setFormatter(formatter)

        root_logger = logging.getLogger("")
        root_logger.addHandler(handler)
        if not hasattr(logging, self.log_level.upper()):
            raise BadConfigKeyError("log_level is set to unknown value (%s)" % self.log_level)
        root_logger.setLevel(getattr(logging, self.log_level.upper()))

        log.defaultObserver.stop()
        observer = log.PythonLoggingObserver()
        observer.start()

    def get_logging_formatter(self):
        import logging
        return logging.Formatter(datefmt='%Y-%m-%d %H:%M:%S',
                                 fmt='%(asctime)s %(levelname)-8s %(message)s')

config = Config()
