#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------
import os
from os import path

from twisted.internet import reactor, defer, task

from mpath_management.utils import dict_intersect, dict_difference, load_file, save_file
from mpath_management.utils import log as logging

import model
import log_msg_id
from config import config
from xml_serialization import XMLSerializer, XMLDeserializer
from native_mpio import NativeMPIONotReadyError

log = logging.Logger("Manager")
event_log = logging.Logger("Events")

class ManagerError(Exception):
    def __init__(self, msg=""):
        Exception.__init__(self, msg)

class WaitingForSnapshotError(ManagerError):
    def __init__(self, msg="waiting for initial snapshot"):
        ManagerError.__init__(self, msg)

class DeviceNotFoundError(ManagerError):
    def __init__(self, device_id):
        ManagerError.__init__(self, "device %s not found" % device_id)

class UnsupportedOperationError(ManagerError):
    def __init__(self, device_id):
        ManagerError.__init__(self, "operation not supported for device %s" % device_id)

class Manager:
    def __init__(self, native_mpio, baseline=None):
        self.native_mpio = native_mpio
        self.baseline = self._load_baseline() if baseline is None else baseline
        self.latest_snapshot = None
        self._start_snapshot_polling()

    def status(self):
        if self.latest_snapshot is None:
            raise WaitingForSnapshotError()
        baseline = self.baseline.copy()
        latest_snapshot = self.latest_snapshot.copy()
        return (baseline, latest_snapshot)

    def update_baseline(self):
        if self.latest_snapshot is None:
            raise WaitingForSnapshotError()
        
        log.info("updating entire baseline to latest snapshot")
        self.baseline = self.latest_snapshot
        self._save_baseline()
    
    def update_baseline_for_device(self, scsi_id):
        if self.latest_snapshot is None:
            raise WaitingForSnapshotError()

        if scsi_id not in self.latest_snapshot.devices \
           and scsi_id not in self.baseline.devices:
            raise DeviceNotFoundError(scsi_id)
        
        log.info("updating baseline for device: %s" % scsi_id)
        if scsi_id in self.latest_snapshot.devices:
            self.baseline.devices[scsi_id] = \
                                           self.latest_snapshot.devices[scsi_id]
        else:
             del self.baseline.devices[scsi_id]
        self.baseline.timestamp = self.latest_snapshot.timestamp
        self._save_baseline()
        
    def set_lb_policy(self, scsi_id, policy, path_params):
        device = self.find_device_from_snapshot_by_scsi_id(scsi_id)
        dev_file_path = device.device_file_path
        def _cb(path_settings):
            device = self.find_device_from_snapshot_by_scsi_id(scsi_id)
            device.lb_policy = policy
            for id, settings in path_settings.items():
                if 'primary' in settings:
                    device.device_paths[id].primary = settings['primary']
                if 'weight' in settings:
                    device.device_paths[id].weight = settings['weight']
        d = defer.maybeDeferred(self.native_mpio.set_load_balancing_policy,
                                dev_file_path, policy, path_params)
        d.addCallback(_cb)
        return d

    def find_device_from_snapshot_by_scsi_id(self, scsi_id):
        if self.latest_snapshot is None:
            raise WaitingForSnapshotError()

        if scsi_id not in self.latest_snapshot.devices:
            raise DeviceNotFoundError(scsi_id)
        
        return self.latest_snapshot.devices[scsi_id]

    def on_new_snapshot(self, cur_snapshot):
        prev_snapshot = self.latest_snapshot
        self.latest_snapshot = cur_snapshot
        if self.baseline is not None:
            self._update_baseline(prev_snapshot, cur_snapshot)
            if prev_snapshot is not None:
                self._notify_event_log(prev_snapshot, cur_snapshot)
        else:
            self.baseline = cur_snapshot
            self.latest_snapshot = cur_snapshot
            self._save_baseline()

    def _start_snapshot_polling(self):
        log.debug("starting snapshot polling")
        self.snapshot_polling_task = task.LoopingCall(self._query_snapshot)
        self.snapshot_polling_task.start(config.snapshot_polling_interval)

    def _query_snapshot(self):
        log.debug("querying for a new snapshot")
        d = defer.maybeDeferred(self.native_mpio.query_snapshot)
        def _eb(failure):
            try:
                failure.trap(NativeMPIONotReadyError)
            except:
                event_log.info("Querying the operating system for device status has failed. Should this problem persist please contact technical support.",
                         log_msg_id=log_msg_id.FAILED_NATIVE_MPIO_QUERY)
            return None
        d.addCallbacks(self.on_new_snapshot, _eb)
        return d

    def _notify_event_log(self, prev_snapshot, cur_snapshot):
        snapshot_diff = self._calc_snapshot_diff(prev_snapshot, cur_snapshot)

        for dev in snapshot_diff['removed_devices']:
            self._on_device_down(dev)

        for dev in snapshot_diff['new_devices']:
            self._on_device_up(dev)

        for dev_id, path_dict in snapshot_diff['removed_paths'].items():
            cur_dev = cur_snapshot.devices[dev_id]
            baseline_count = len(self.baseline.devices[dev_id].device_paths)
            snapshot_count = len(cur_snapshot.devices[dev_id].device_paths)
            for path in path_dict.values():
                self._on_device_path_down(cur_dev, path, baseline_count,
                                          snapshot_count)

        for dev_id, path_dict in snapshot_diff['new_paths'].items():
            cur_dev = cur_snapshot.devices[dev_id]
            for path in path_dict.values():
                self._on_device_path_up(cur_dev, path)

    def _update_baseline(self, prev_snapshot, cur_snapshot):
        baseline_diff = self._calc_snapshot_diff(self.baseline, cur_snapshot)
        
        new_baseline = model.Snapshot(timestamp=cur_snapshot.timestamp)
        baseline_modified = False
        
        # Add all new devices to the baseline and report it.
        for dev in baseline_diff['new_devices']:
            new_baseline.devices[dev.scsi_address] = dev.copy()
            baseline_modified = True

        # Here's the tricky part: the path IDs of the current snapshot may not
        # necessarily correspond to the path IDs of the baseline, even if the
        # count is the same. This scenario may happen if there was a disconnect
        # at some level (switch, driver, etc.) and the new path allocated got
        # a new ID - no one assures us that the ID is the same (especially
        # across reboots for example).
        # To overcome this, we try to estimate which paths in the current
        # snapshot correspond to paths in the baseline by correlating the paths
        # with the SCSI LUN and target.
        for baseline_dev in baseline_diff['existing_devices']:
            cur_dev = cur_snapshot.devices[baseline_dev.scsi_address]
            cur_dev_paths = cur_dev.device_paths
            baseline_dev_paths = baseline_dev.device_paths
            baseline_path_count = len(baseline_dev_paths)
            snapshot_path_count = len(cur_dev_paths)

            if baseline_path_count < snapshot_path_count or \
               (baseline_path_count == snapshot_path_count and \
                   baseline_dev_paths.keys() != cur_dev_paths.keys()):
                # If we have more paths in the current snapshot or we have the
                # same path count but different IDs, we copy the new snapshot
                # data.
                new_baseline.devices[cur_dev.scsi_address] = cur_dev.copy()
                baseline_modified = True
            elif baseline_path_count > snapshot_path_count:
                # We lost some paths. Theoretically, all the path IDs could have
                # changed in the snapshot, so we try here to find paths that
                # were in the baseline but not in the new snapshot.
                new_baseline.devices[cur_dev.scsi_address] = cur_dev.copy()

                # First, see if any of the current paths are new.
                for id in cur_dev_paths.keys():
                    if id not in baseline_dev_paths:
                        baseline_modified = True
                        break

                # Now, try and figure out which paths in the baseline least
                # likely to be in the current snapshot, and these are the ones
                # we keep in our baseline.
                path_score = self._calc_path_correlation(cur_dev_paths,
                                                         baseline_dev_paths)
                ids_by_score = self._path_ids_by_score(path_score)
                d = baseline_path_count - snapshot_path_count

                for id in ids_by_score[0:d]:
                    new_baseline.devices[cur_dev.scsi_address].device_paths[id] = baseline_dev_paths[id].copy()
            else:
                new_baseline.devices[cur_dev.scsi_address] = baseline_dev.copy()
        
        if baseline_modified:
            self.baseline = new_baseline
            self._save_baseline()

    def _calc_path_correlation(self, a, b):
        def calc_key(path):
            return "%s_%s" % (path.scsi_lun, path.scsi_target)
        
        map_by_key = dict()
        for path in a.values():
            key = calc_key(path)
            if key in map_by_key:
                map_by_key[key].append(path)
            else:
                map_by_key[key] = [ path ]

        path_score = dict()
        for id, path in b.items():
            if id in a:
                path_score[id] = 1.0 # exact match
            else:
                key = calc_key(path)
                paths = map_by_key.get(key, [])
                if len(a) > 0:
                    path_score[id] = float(len(paths)) / len(a)
                else:
                    path_score[id] = 0.0

        return path_score

    def _path_ids_by_score(self, path_score, reverse=False):
        return map(lambda x: x[0],
                   sorted(path_score.iteritems(), key=lambda x: x[1],
                          reverse=reverse))

    def _on_device_up(self, device):
        event_log.info("New device found:\n%s" % device.summary(),
                       log_msg_id=log_msg_id.NEW_DEVICE_FOUND)

    def _on_device_down(self, device):
        event_log.warn("Device removed:\n%s" % device.summary(),
                       log_msg_id=log_msg_id.DEVICE_REMOVED)

    def _on_device_path_up(self, device, path):
        event_log.info("New path found for device %s (%s):\n%s" %
            (device.scsi_address, device.device_file_path, path.summary()),
            log_msg_id=log_msg_id.NEW_PATH_FOUND)

    def _on_device_path_down(self, cur_dev, path, baseline_count, cur_count):
        if cur_count * 2 <= baseline_count:
            level = logging.CRITICAL
        else:
            level = logging.WARNING
        event_log.message("Path removed for device %s (%s):\n%s" %
            (cur_dev.scsi_address, cur_dev.device_file_path, path.summary()),
            logLevel=level, log_msg_id=log_msg_id.PATH_REMOVED)
        
    def _calc_snapshot_diff(self, old_snap, new_snap):
        """
        Calculates the difference between two snapshots and returns a dict with:
          new_devices - list of devices that weren't in the old snapshot
          removed_devices - list of devices that aren't in the new snapshot
          existing_devices - list of devices that are in both snapshots
          new_paths - list of paths that weren't in the old snapshot
          removed_paths - list of paths that aren't in the new snapshot
        """
        new_devices = dict_difference(new_snap.devices,
                                      old_snap.devices).values()
        removed_devices = dict_difference(old_snap.devices,
                                          new_snap.devices).values()
        existing_devices = dict_intersect(old_snap.devices,
                                          new_snap.devices).values()
        new_paths = {}
        removed_paths = {}

        for old_dev in existing_devices:
            new_dev = new_snap.devices[old_dev.scsi_address]
            new_paths[old_dev.scsi_address] = \
                dict_difference(new_dev.device_paths, old_dev.device_paths)
            removed_paths[old_dev.scsi_address] = \
                dict_difference(old_dev.device_paths, new_dev.device_paths)
        return dict(new_devices=new_devices, removed_devices=removed_devices,
                    existing_devices=existing_devices, new_paths=new_paths,
                    removed_paths=removed_paths)

    def _load_baseline(self):
        try:
            baseline_file_name = self._baseline_file_name()
            if path.exists(baseline_file_name):
                log.info("%s exists, loading baseline from disk." %
                         baseline_file_name)
                data = load_file(baseline_file_name)
                return XMLDeserializer().deserialize(data)
            else:
                log.info("%s doesn't exist, not loading baseline from disk." %
                         baseline_file_name)
        except Exception, e:
            log.warn("failed loading baseline file", exc_info=e,
                     log_msg_id=log_msg_id.FAILED_LOADING_BASELINE)
        return None

    def _save_baseline(self):
        try:
            save_file(self._baseline_file_name(),
                      XMLSerializer().serialize(self.baseline))
        except Exception, e:
            log.warn("failed saving baseline file %s" %
                     self._baseline_file_name(), exc_info=e,
                     log_msg_id=log_msg_id.FAILED_SAVING_BASELINE)

    def _baseline_file_name(self):
        return config.get_path_from_data_dir("baseline.xml")
