#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program  
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
# 
#-------------------------------------------------------------------------------------------------
import time
import copy
from cStringIO import StringIO

class Snapshot(object):
    def __init__(self, timestamp=None, devices=None):
        object.__init__(self)
        self.timestamp = timestamp or time.time()
        self.devices = devices or {}

    def copy(self):
        return copy.deepcopy(self)

    def to_str(self, indent=0):
        out = StringIO()
        out.write("%sSnapshot timestamp=%s devices=%d\n" % (' ' * indent,
                    self.timestamp, len(self.devices)))
        for device in self.devices.values():
            out.write(device.to_str(indent=indent+2))
        content = out.getvalue()
        out.close()
        return content
        
    def __str__(self):
        return self.to_str()

class Device(object):
    # Properties we can get from XSF:
    # name --> device_file_path? (e.g. '\\.\PHYSICALDRIVE1')
    # scsi_serial: this is the unique identifier.
    # scsi_standard_inquiry: collection that includes all page83 info
    #  specifically: vendor, product
    # mount_point
    # children (paths)

    # For each child:
    #  iSCSI:
    #    scsi_host (can be null)
    #    lun
    #    target_id
    LoadBalancingPolicy = dict(FailOverOnly=1,
                               RoundRobin=2,
                               RoundRobinWithSubset=3,
                               DynamicLeastQueueDepth=4,
                               WeightedPaths=5,
                               LeastBlocks=6)

    # Note: scsi_address must _uniquely_ identify a device, regardless of
    # the mount point, reboots or removal/insertion.
    def __init__(self, scsi_address=None, scsi_vendor_id=None,
                 scsi_product_id=None, scsi_product_revision=None,
                 lb_policy=None, device_file_path=None, mount_points=None,
                 device_paths=None, os_info=None):
        object.__init__(self)
        self.scsi_address = scsi_address
        self.scsi_vendor_id = scsi_vendor_id
        self.scsi_product_id = scsi_product_id
        self.scsi_product_revision = scsi_product_revision
        self.lb_policy = lb_policy
        self.device_file_path = device_file_path
        self.mount_points = mount_points or []
        self.device_paths = device_paths or {}
        self.os_info = os_info or {}
        self.supported_lb_policies = []

    def copy(self):
        return copy.deepcopy(self)
    
    def to_str(self, indent=0):
        out = StringIO()
        out.write("%sDevice:\n" % (' ' * indent))
        out.write(self._to_str_device_attributes(indent + 2))
        out.write("\n")
        out.write(self._to_str_device_paths(indent + 2))
        out.write("\n")
        content = out.getvalue()
        out.close()
        return content

    def summary(self):
        return self._to_str_device_attributes()

    def _to_str_device_attributes(self, indent=0):
        lines = []
        lines.append("Number of paths: %d" % len(self.device_paths))
        lines.append("Device file path: %s" % self.device_file_path)
        lines.append("SCSI address: %s" % self.scsi_address)
        lines.append("Vendor: %s" % self.scsi_vendor_id)
        lines.append("Product: %s" % self.scsi_product_id)
        lines.append("Product Revision: %s" % self.scsi_product_revision)
        lines.append("Load balancing policy: %s" % self.lb_policy)
        if len(self.mount_points) == 1:
            lines.append("Mount point: %s" % self.mount_points[0])
        elif len(self.mount_points) > 1:
            lines.append("Mount points:")
            for mount_point in self.mount_points:
                lines.append("%s%s" % (' ' * 4, mount_point))
        #lines.append("OS specific: %s" % self.os_info)
        return "\n".join([ "%s%s" % (' ' * indent, line) for line in lines ])
        
    def _to_str_device_paths(self, indent=0):
        lines = []
        lines.append("%sPaths (%d):" % (' ' * indent, len(self.device_paths)))
        for path in self.device_paths.values():
            lines.append(path.to_str(indent + 2))
        return "\n".join(lines)

    def __cmp__(self, other):
        return cmp(self.scsi_address, other.scsi_address)

    def __str__(self):
        return self.to_str()

class XIVDevice(Device):
    # TPGS: xiv_vol_name, xiv_host_name
    def __init__(self, xiv_machine_serial=None, xiv_vol_serial=None,
                 xiv_vol_name=None, xiv_host_name=None, *args, **kwargs):
        Device.__init__(self, *args, **kwargs)
        self.xiv_machine_serial = xiv_machine_serial
        self.xiv_vol_serial = xiv_vol_serial
        self.xiv_vol_name = xiv_vol_name
        self.xiv_host_name = xiv_host_name

    def to_str(self, indent=0):
        out = StringIO()
        out.write("%sXIV Device:\n" % (' ' * indent))
        out.write(self._to_str_device_attributes(indent + 2))
        out.write("\n")
        out.write(self._to_str_device_paths(indent + 2))
        out.write("\n")
        content = out.getvalue()
        out.close()
        return content

    def _to_str_device_attributes(self, indent=0):
        return "\n".join((Device._to_str_device_attributes(self, indent),
                         self._to_str_xiv_device_attributes(indent)))
    
    def _to_str_xiv_device_attributes(self, indent=0):
        lines = []
        lines.append("XIV machine serial number: %s" % self.xiv_machine_serial)
        lines.append("XIV volume name: %s" % self.xiv_vol_name)
        lines.append("XIV volume serial number: %s" % self.xiv_vol_serial)
        lines.append("XIV host name: %s" % self.xiv_host_name)
        return "\n".join([ "%s%s" % (' ' * indent, line) for line in lines ])
        
class DevicePath(object):
    InterfaceType = dict(iSCSI=1, FiberChannel=2)

    def __init__(self, id=None, interface_type=None,
                 scsi_target=None, scsi_lun=None,
                 io_bytes_read=None, io_bytes_written=None,
                 io_read_ops=None, io_write_ops=None,
                 io_errors=None, io_retries=None, primary=None, weight=None):
        object.__init__(self)
        self.id = id
        self.interface_type = interface_type
        self.scsi_target = scsi_target
        self.scsi_lun = scsi_lun
        self.io_bytes_read = io_bytes_read
        self.io_bytes_written = io_bytes_written
        self.io_read_ops = io_read_ops
        self.io_write_ops = io_write_ops
        self.io_errors = io_errors
        self.io_retries = io_retries
        self.primary = primary
        self.weight = weight

    def copy(self):
        return copy.deepcopy(self)

    def to_str(self, indent=0):
        out = StringIO()
        out.write("%sPath:\n" % (' ' * indent,))
        out.write(self._to_str_path_attributes(indent + 2))
        content = out.getvalue()
        out.close()
        return content

    def __cmp__(self, other):
        return cmp(self.id, other.id)

    def __str__(self):
        return self.to_str()

    def summary(self):
        return self._to_str_path_attributes()

    def _to_str_path_attributes(self, indent=0):
        lines = []
        lines.append("ID: %s" % self.id)
        lines.append("Interface type: %s" % self.interface_type)
        lines.append("SCSI target: %s LUN: %s" % (self.scsi_target,
                                                  self.scsi_lun))
        if self.io_bytes_read is not None:
            lines.append("Number of bytes read: %d" % self.io_bytes_read) 
        if self.io_bytes_written is not None:
            lines.append("Number of bytes written: %d" % self.io_bytes_written)
        if self.io_read_ops is not None:
            lines.append("Number of read operations: %d" % self.io_read_ops)
        if self.io_write_ops is not None:
            lines.append("Number of write operations: %d" % self.io_write_ops)
        if self.io_errors is not None:
            lines.append("Number of I/O errors: %d" % self.io_errors)
        if self.io_retries is not None:
            lines.append("Number of I/O retries: %d" % self.io_retries)
        if self.primary is not None:
            lines.append("Is primary path: %s" % self.primary)
        if self.weight is not None:
            lines.append("Weight of path: %d" % self.weight)
        return "\n".join([ "%s%s" % (' ' * indent, line) for line in lines ])

class XIVDevicePath(DevicePath):
    def __init__(self, machine_module=None, machine_port=None,
                 machine_rack=None, *args, **kwargs):
        DevicePath.__init__(self, *args, **kwargs)
        self.machine_module = machine_module
        self.machine_port = machine_port
        self.machine_rack = machine_rack

    def to_str(self, indent=0):
        out = StringIO()
        out.write("%sXIV Path:\n" % (' ' * indent,))
        out.write(self._to_str_path_attributes(indent + 2))
        out.write("\n")
        out.write(self._to_str_xiv_path_attributes(indent + 2))
        content = out.getvalue()
        out.close()
        return content

    def _to_str_path_attributes(self, indent=0):
        return "\n".join((DevicePath._to_str_path_attributes(self, indent),
                         self._to_str_xiv_path_attributes(indent)))
        
    def _to_str_xiv_path_attributes(self, indent):
        lines = []
        lines.append("XIV machine module: %s" % self.machine_module)
        lines.append("XIV machine port: %s" % self.machine_port)
        lines.append("XIV machine rack: %s" % self.machine_rack)
        return "\n".join([ "%s%s" % (' ' * indent, line) for line in lines ])
