#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------
import time
import re
from cStringIO import StringIO

from twisted.internet import reactor, defer
from twisted.web import server, resource, http

from mpath_management.utils.log import Logger
from .xml_serialization import XMLSerializer
from .manager import WaitingForSnapshotError, DeviceNotFoundError
from .manager import UnsupportedOperationError
import model
import log_msg_id
from config import config

import random
import struct
import base64
import hashlib
from Crypto.Cipher import DES
from hak.common.definitions import HOSTATTACHMENT_VERSION

log = Logger("http")
event_log = Logger("Events")

PROTOCOL_VERSION = "1.1"
AGENT_VERSION = HOSTATTACHMENT_VERSION

# 120 seconds, should take no more than a sec in real life, but we're allowing
# more in case there's high load on the agent.
CHALLENGE_MAX_VALID_TIME = 120

def random_string(length):
    s = []
    for i in xrange(0, length):
        s.append(random.randrange(256))
    return struct.pack("%dB" % length, *s)

# Pads with NULLs at the end, so don't use NULLs in the data.
def pad_and_encrypt_and_b64encode(cipher, data):
    r = len(data) % cipher.block_size
    if r != 0:
        data += "\x00" * (cipher.block_size - r)
    return base64.b64encode(cipher.encrypt(data))

def b64decode_and_decrypt_and_unpad(cipher, data):
    encrypted_string = base64.b64decode(data)
    decrypted_string = cipher.decrypt(encrypted_string)
    return decrypted_string.rstrip("\x00")

class RequestError(Exception):
    def http_code(self):
        return getattr(http, self.name)

    def http_status_line(self):
        return http.RESPONSES[self.http_code()]

    def __str__(self):
        return ' '.join(self.args) if len(self.args) > 0 else self.http_status_line()

class ForbiddenRequestError(RequestError):
    name = 'FORBIDDEN'

class ServiceUnavailableError(RequestError):
    name = 'SERVICE_UNAVAILABLE'

class BadRequestError(RequestError):
    name = 'BAD_REQUEST'

class NotFoundError(RequestError):
    name = 'NOT_FOUND'

class ServiceNotImplementedError(RequestError):
    name = 'NOT_IMPLEMENTED'

class ConfigurableWebResource(resource.Resource):
    isLeaf = True

    def __init__(self):
        resource.Resource.__init__(self)

    def render_GET(self, request):
        return self._handle_request(request)

    def render_POST(self, request):
        return self._handle_request(request)

    def render_PUT(self, request):
        return self._handle_request(request)

    def _call_url_mapping_match(self, request):
        for path_re, func in self._get_url_mapping().items():
            # TODO pre-compile all the regexps and use it here.
            regexp = re.compile(path_re)
            match = regexp.match(request.path)
            if match:
                kwargs = match.groupdict()
                pos = set(range(1, regexp.groups)) - \
                      set(regexp.groupindex.values())
                args = match.groups(*pos) if len(pos) > 0 else []
                return func(*args, request=request, **kwargs)
        raise NotFoundError()

    def _handle_request(self, request):
        try:
            tup = self._call_url_mapping_match(request)
        except RequestError, e:
            self._send_error(request, e.http_code(), str(e))
            return server.NOT_DONE_YET
        except Exception, e:
            log.info("caught exception for request %s with parameters %s" % (request, repr(request.args)), exc_info=e)
            event_log.warn("A request for the XIV MPIO Agent has failed. Should this problem persist please contact " +
                           "technical support.", log_msg_id=log_msg_id.FAILED_REQUEST)
            self._send_error(request, http.INTERNAL_SERVER_ERROR, str(e))
            return server.NOT_DONE_YET

        if isinstance(tup, defer.Deferred):
            def _cb(tup):
                self._send_response(request, *tup)
                return None

            def _request_eb(failure):
                if not failure.check(RequestError):
                    return failure
                self._send_error(request, failure.value.http_code(), str(failure.value))
                return None

            def _eb(failure):
                log.info("request %s with parameters %s has failed: %s" % (request, repr(request.args),
                                                                           str(failure.value)), failure=failure)
                event_log.warn("A request for the XIV MPIO Agent has failed. Should this problem persist please " +
                               "contact technical support.", log_msg_id=log_msg_id.FAILED_REQUEST)
                self._send_error(request, http.INTERNAL_SERVER_ERROR,
                                 failure.getErrorMessage())
                return None

            tup.addCallback(_cb)
            tup.addErrback(_request_eb)
            tup.addErrback(_eb)
        else:
            self._send_response(request, *tup)

        return server.NOT_DONE_YET

    def _send_response(self, request, status, content='', headers=None):
        if headers is None:
            headers = {}
        for header in headers:
            request.setHeader(header, headers[header])
        request.setResponseCode(int(status))
        request.setHeader('content-length', str(len(content)))
        request.write(content)
        request.finish()

    def _send_error(self, request, status, content='', headers=None):
        if headers is None:
            headers = {}
        return self._send_response(request, status,
                    "<html><body>Error: %s</body></html>" % content, headers)

    def _get_url_mapping(self):
        return {}

    def _get_user(self, request):
        if config.require_authentication:
            from twisted.web.woven.simpleguard import Authenticated
            return request.getComponent(Authenticated).name
        else:
            return None

    def _abort_unless(self, request, **kwargs):
        if 'method' in kwargs:
            if request.method != kwargs['method']:
                methods = set(('GET', 'POST', 'PUT', 'HEAD', 'DELETE'))
                raise server.UnsupportedMethod(list(methods - set((kwargs['method'],))))
        if 'user' in kwargs and config.require_authentication:
            if self._get_user(request) != kwargs['user']:
                raise ForbiddenRequestError()

    def _sanitize_choice_or_abort(self, request, key_name, choices):
        if key_name not in request.args or request.args[key_name] is None:
            raise BadRequestError("expected parameter %s" % key_name)

        value = request.args[key_name][-1]
        for v in [ value, value.lower() ]:
            if v in choices:
                return v

        raise BadRequestError("parameter %s has unknown value: %s" %
                              (key_name, value))

class AgentWebResource(ConfigurableWebResource):
    def __init__(self, manager):
        ConfigurableWebResource.__init__(self)
        self.manager = manager
        self.xml_serializer = XMLSerializer()
        self.rand_key = random_string(8)

    def _status(self, request):
        self._abort_unless(request, method='GET')
        try:
            baseline, latest_snapshot = self.manager.status()

            output = StringIO()
            output.write("<status>\n")
            output.write("<baseline>\n")
            self.xml_serializer.serialize(baseline, output)
            output.write("</baseline>\n")
            output.write("<latest>\n")
            self.xml_serializer.serialize(latest_snapshot, output)
            output.write("</latest>\n")
            output.write("</status>")
            content = output.getvalue()
            output.close()

            return http.OK, content, { "content-type": "text/xml" }
        except WaitingForSnapshotError:
            raise ServiceUnavailableError("Agent is still acquiring device information")

    def _stats(self, request):
        self._abort_unless(request, method='GET')
        return http.OK, "Stats not available yet.", {}

    def _shutdown(self, request):
        self._abort_unless(request, method='GET', user='manager')
        reactor.stop()
        return http.OK, "Shutting down server, bye bye!", {}

    def _set_lb_policy(self, request, device_id):
        self._abort_unless(request, method='POST', user='manager')
        policy = self._sanitize_choice_or_abort(request, 'policy',
                        set(model.Device.LoadBalancingPolicy.keys()))
        path_params = dict()
        for k, v in request.args.items():
            if k != 'policy':
                path_params[k] = int(v[0])
        def _cb(_):
            return http.OK, "Load balancing policy was set to %s for device %s" % (policy, device_id)
        d = defer.maybeDeferred(self.manager.set_lb_policy, device_id, policy,
                                path_params)

        d.addCallback(_cb)
        d.addErrback(self._on_device_not_found_error)
        d.addErrback(self._on_waiting_for_snapshot_error)
        d.addErrback(self._on_unsupported_operation_error)

        return d

    def _set_baseline(self, request, device_id=None):
        self._abort_unless(request, method='GET', user='manager')

        if device_id is None:
            d = defer.maybeDeferred(self.manager.update_baseline)
        else:
            d = defer.maybeDeferred(self.manager.update_baseline_for_device,
                                    device_id)
        def _cb(_):
            return http.OK, self.xml_serializer.serialize(self.manager.baseline), { "content-type": "text/xml" }

        d.addCallback(_cb)
        d.addErrback(self._on_device_not_found_error)
        d.addErrback(self._on_waiting_for_snapshot_error)

        return d

    def _set_password(self, request):
        self._abort_unless(request, method='GET')
        op = self._sanitize_choice_or_abort(request, 'op',
                                            [ 'challenge', 'set' ])

        cur_user = self._get_user(request) or 'manager'
        if 'user' in request.args:
            user = self._sanitize_choice_or_abort(request, 'user',
                                                  [ 'monitor', 'manager' ])
            # Allow only manager to set monitor's password.
            if cur_user != 'manager' and cur_user != user:
                log.info("set_password denied: user %s cannot modify user %s's password" % (cur_user, user))
                raise ForbiddenRequestError()
        else:
            user = cur_user

        now = time.time()
        if op == 'challenge':
            challenge = self._create_password_challenge(now)
            return http.OK, challenge, { "content-type": "text/ascii" }
        elif op == 'set':
            if 'challenge' not in request.args or 'val' not in request.args:
                raise BadRequestError("expected parameter challenge and val")

            challenge = request.args['challenge'][-1]
            val = request.args['val'][-1]

            try:
                self._set_password_with_challenge(cur_user, user, challenge,
                                                  val, now)
                return http.OK, "New password has been set."
            except (TypeError, ValueError), e:
                log.debug("caught set_password error", exc_info=e)
                raise BadRequestError("malformatted challenge or val")
        raise BadRequestError("expected op and user")

    def _create_password_challenge(self, now):
        salt = random_string(4)
        des = DES.new(self.rand_key, DES.MODE_CBC)
        return pad_and_encrypt_and_b64encode(des, salt + str(now))

    def _set_password_with_challenge(self, cur_user, user, challenge, val, now):
        des = DES.new(self.rand_key, DES.MODE_CBC)
        decoded_challenge = \
            b64decode_and_decrypt_and_unpad(des, challenge)

        if len(decoded_challenge) < 5:
            raise ValueError("challenge length is too short")

        prev_timestamp = float(decoded_challenge[4:])
        log.debug("_set_password_with_challenge cur_time=%d challenge_time=%d" % (now, prev_timestamp))
        if abs(now - prev_timestamp) > CHALLENGE_MAX_VALID_TIME:
            raise ValueError("challenge is out of the time window")

        # Challenge is valid, now let's try to decrypt the password.
        # DES key first 8 bytes of
        # sha1(older user password + user + challenge)
        config_old_password_key = "%s_password" % cur_user
        config_new_password_key = "%s_password" % user
        old_password = getattr(config, config_old_password_key)
        m = hashlib.sha1()
        m.update(old_password + user + challenge)
        key = m.digest()[0:8]
        des = DES.new(key, DES.MODE_CBC)
        result = b64decode_and_decrypt_and_unpad(des, val)

        if not result.startswith(old_password):
            log.info("someone tried to set a new password for user %s with an incorrect old password (challenge: %s)" %
                     (user, challenge))
            raise ValueError

        config.set_value(config_new_password_key,
                         result[len(old_password):])

    def _capabilities(self, request):
        self._abort_unless(request, method='GET')

        out = StringIO()
        out.write("<capabilities protocol_version=\"%s\" agent_version=\"%s\">\n" % (PROTOCOL_VERSION, AGENT_VERSION))

        from xml.sax.saxutils import quoteattr
        uname = []
        platform_uname = self._get_machine_info()
        for key in ('platform', 'system', 'node', 'release', 'version',
                    'machine', 'processor'):
            if key in platform_uname and len(platform_uname[key]) > 0:
                uname.append("%s=%s" % (key, quoteattr(platform_uname[key])))
        out.write("<platform %s/>" % " ".join(uname))
        if (not config.require_authentication) or \
               self._get_user(request) == 'manager':
            out.write("<set_baseline/>\n")
            out.write("<set_lb_policy>\n")
            for policy in self._get_available_lb_policies():
                out.write("<policy>%s</policy>\n" % policy)
            out.write("</set_lb_policy>\n")
        out.write("<status/>")
        out.write("</capabilities>")
        content = out.getvalue()
        out.close()

        return http.OK, content, { "content-type": "text/xml" }

    def _get_machine_info(self):
        """Return a dict with Python's platform.uname() compatible keys
        (system, node, release, version, machine, processor)"""
        import platform
        uname = platform.uname()

        return dict(system=uname[0], node=uname[1], release=uname[2],
                    version=uname[3], machine=uname[4],
                    processor=uname[5])

    def _get_available_lb_policies(self):
        return model.Device.LoadBalancingPolicy.keys()

    def _on_device_not_found_error(self, failure):
        failure.trap(DeviceNotFoundError)
        raise NotFoundError()

    def _on_waiting_for_snapshot_error(self, failure):
        failure.trap(WaitingForSnapshotError)
        raise ServiceUnavailableError()

    def _on_unsupported_operation_error(self, failure):
        failure.trap(UnsupportedOperationError)
        raise ServiceNotImplementedError()

    def _get_url_mapping(self):
        return {
            '/status': self._status,
            '/stats': self._stats,
            '/shutdown': self._shutdown,
            r'/device/(?P<device_id>.*)/set_lb_policy': self._set_lb_policy,
            r'/device/(?P<device_id>.*)/set_baseline': self._set_baseline,
            '/set_baseline': self._set_baseline,
            r'/set_password': self._set_password,
            '/capabilities': self._capabilities,
        }