#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------
import cStringIO
import time
from xml.sax.saxutils import quoteattr
from xml.sax.handler import ContentHandler

from mpath_management.utils.text import camel_case_to_wide_name
from mpath_management.utils.log import Logger
import model

log = Logger("XMLSerializer")

PATH_ATTRIBUTES = 'id interface_type scsi_target scsi_lun io_bytes_read io_bytes_written io_read_ops'.split() + \
                                                            'io_write_ops io_errors io_retries primary weight'.split()
XIV_PATH_ATTRIBUTES = PATH_ATTRIBUTES + 'machine_module machine_port machine_rack'.split()
DEVICE_ATTRIBUTES = 'scsi_address scsi_vendor_id scsi_product_id scsi_product_revision'.split() + \
                                                                                'lb_policy device_file_path'.split()
XIV_DEVICE_ATTRIBUTES = DEVICE_ATTRIBUTES + \
                        'xiv_machine_serial xiv_vol_serial xiv_vol_name xiv_host_name'.split()
                        
class XMLSerializer(object):
    def serialize(self, serialization_object, output=None):
        if serialization_object is None:
            raise ValueError("serialize received a None serialization_object")

        class_name = camel_case_to_wide_name(serialization_object.__class__.__name__)
        if class_name.startswith("jelly_"):
            class_name = class_name[len("jelly_"):]
        func_name = "_serialize_" + class_name
                    
        try:
            func = getattr(self, func_name)
        except AttributeError:
            raise ValueError("no XML serialization was defined for %s (object: %s)" % (func_name, serialization_object))

        buf = output or cStringIO.StringIO()
        func(serialization_object, buf)
        content = buf.getvalue()
        if output is None:
            buf.close()
        return content

    def _serialize_snapshot(self, serialization_object, buf):
        buf.write('<snapshot timestamp="%s">\n' % serialization_object.timestamp)
        for device in serialization_object.devices.values():
            self.serialize(device, buf)
        buf.write('</snapshot>\n')

    _serialize_receiver_snapshot = _serialize_snapshot

    def _serialize_device_elements(self, serialization_object, buf):
        for mount_point in serialization_object.mount_points:
            buf.write('<mount_point path=%s />' % quoteattr(mount_point))
        for path in serialization_object.device_paths.values():
            self.serialize(path, buf)
        if len(serialization_object.supported_lb_policies) > 0:
            buf.write("<supported_lb_policies>")
            for lb_policy in serialization_object.supported_lb_policies:
                buf.write("<policy>%s</policy>" % lb_policy)
            buf.write("</supported_lb_policies>")
        
    def _serialize_device(self, serialization_object, buf):
        buf.write('<device %s>' % self._serialize_attributes(serialization_object, DEVICE_ATTRIBUTES))
        self._serialize_device_elements(serialization_object, buf)
        buf.write("</device>")

    def _serialize_xiv_device(self, serialization_object, buf):
        buf.write('<xiv_device %s>' % self._serialize_attributes(serialization_object, XIV_DEVICE_ATTRIBUTES))
        self._serialize_device_elements(serialization_object, buf)
        buf.write("</xiv_device>")
        
    def _serialize_device_path(self, serialization_object, buf):
        buf.write('<path %s/>' % self._serialize_attributes(serialization_object, PATH_ATTRIBUTES))
    
    def _serialize_xiv_device_path(self, serialization_object, buf):
        buf.write('<xiv_path %s/>' % self._serialize_attributes(serialization_object, XIV_PATH_ATTRIBUTES))

    def _serialize_attributes(self, serialization_object, attributes):
        out = []
        for attr in attributes:
            val = getattr(serialization_object, attr)
            if val is not None:
                out.append("%s=%s" % (attr, quoteattr(str(val))))
        return " ".join(out)

class XMLDeserializer:
    class UnknownElementError(ValueError):
        def __init__(self, name):
            ValueError.__init__(self, "unknown element %s" % name)

    class BadNestingError(ValueError):
        def __init__(self, name):
            ValueError.__init__(self, "can't handled nested %s elements" % name)

    class BadParentError(ValueError):
        def __init__(self, child, parent):
            ValueError.__init__(self, "can't define child %s outside of parent %s" % (child, parent))

    class BadEndError(ValueError):
        def __init__(self, name):
            ValueError.__init__(self, "can't close a %s element without a start" % name)

    class Handler(ContentHandler):
        def startDocument(self):
            self.entity = None
            self.snapshot = None
            self.device = None
            self.mount_point = None
            self.path = None

        def startElement(self, name, attrs):
            func = getattr(self, "_start_" + name)
            if func is None:
                raise self.UnknownElementError(name)
            func(attrs)
            
        def endElement(self, name):
            func = getattr(self, "_end_" + name)
            if func is None:
                raise self.UnknownElementError(name)
            func()

        def _start_snapshot(self, attrs):
            if self.snapshot is not None:
                raise self.BadNestingError("snapshot")
            f = float(attrs['timestamp'])
            t = time.mktime(time.localtime(f))
            t += (f - int(f))
            self.snapshot = model.Snapshot(timestamp=t)
        
        def _end_snapshot(self):
            if self.snapshot is None:
                raise self.BadEndError("snapshot")
            self.entity = self.snapshot
            self.snapshot = None

        def _start_device(self, attrs):
            if self.device is not None:
                raise self.BadNestingError("device")
            if self.snapshot is None:
                raise self.BadParentError("device", "snapshot")

            self.device = model.Device()
            self._set_attributes(self.device, attrs)

        def _end_device(self):
            if self.device is None:
                raise self.BadEndError("device or xiv_device")
            self.snapshot.devices[self.device.scsi_address] = self.device
            self.device = None

        def _start_xiv_device(self, attrs):
            if self.device is not None:
                raise self.BadNestingError("xiv_device")
            if self.snapshot is None:
                raise self.BadParentError("xiv_device", "snapshot")
            self.device = model.XIVDevice()
            self._set_attributes(self.device, attrs)
        
        def _end_xiv_device(self):
            self._end_device()

        def _start_mount_point(self, attrs):
            if self.mount_point is not None:
                raise self.BadNestingError("mount_point")
            if self.device is None:
                raise self.BadParentError("mount_point", "device")
            self.mount_point = attrs["path"]

        def _end_mount_point(self):
            if self.mount_point is None:
                raise self.BadEndError("mount_point")
            self.device.mount_points.append(self.mount_point)
            self.mount_point = None

        def _start_path(self, attrs):
            if self.path is not None:
                raise self.BadNestingError("path")
            if self.device is None:
                raise self.BadParentError("path", "device")
            self.path = model.DevicePath()
            self._set_attributes(self.path, attrs)

        def _end_path(self):
            if self.path is None:
                raise self.BadEndError("path or xiv_path")
            for attr in 'io_bytes_read io_bytes_written io_read_ops io_write_ops io_errors io_retries weight'.split():
                if getattr(self.path, attr, None) is not None:
                    setattr(self.path, attr, int(getattr(self.path, attr)))
            if getattr(self.path, 'primary', None) is not None:
                setattr(self.path, 'primary', getattr(self.path, 'primary').lower() == 'true')
            self.device.device_paths[self.path.id] = self.path
            self.path = None

        def _start_xiv_path(self, attrs):
            if self.path is not None:
                raise self.BadNestingError("xiv_path")
            if self.device is None:
                raise self.BadParentError("xiv_path", "device")
            self.path = model.XIVDevicePath()
            self._set_attributes(self.path, attrs)

        def _end_xiv_path(self):
            self._end_path()
            for attr in 'machine_module machine_port machine_rack'.split():
                if getattr(self.path, attr, None) is not None:
                    setattr(self.path, attr, int(getattr(self.path, attr)))

        def _start_supported_lb_policies(self, attrs):
            pass
        def _end_supported_lb_policies(self):
            pass
        def _start_policy(self, attrs):
            pass
        def _end_policy(self):
            pass

        def _set_attributes(self, serialization_object, attrs):
            for attr in attrs.keys():
                setattr(serialization_object, attr, attrs[attr])
    
    def deserialize(self, string):
        from xml.sax import parseString
        handler = XMLDeserializer.Handler()
        parseString(string, handler)
        return handler.entity
