#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------
import logging
import traceback
from cStringIO import StringIO

LOG_LEVEL_TO_NAME = {
    logging.DEBUG: 'DEBUG',
    logging.INFO: 'INFO',
    logging.WARNING: 'WARNING',
    logging.ERROR: 'ERROR',
    logging.CRITICAL: 'CRITICAL'
}

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

REPEATABLE_MESSAGES_PRUNE_TIME = 10 # every 10 seconds

PERIOD_ONCE = 86400 * 365 * 10 # 10 years --> one message per boot :)

# Note: NOT thread-safe!
class Logger:
    def __init__(self, name):
        """Lazily initialize the logger. This is done so the following pattern
        can be used:

        <inside a Python module>
        import utils.log as logging
        log = logging.Logger("my_module")

        ...
        def foo():
            log.info("foo called")
        ...
        </end inside a python module>

        <inside main>
        import my_module # logging isn't initialized yet!

        # configure logging (handlers, etc.)
        ...
        my_module.foo()  # now it's working
        """
        self.name = name
        self.log = None
        self.repeatable_messages = {}
        self.last_prune_time = 0
        
    def debug(self, *args, **kwargs):
        self.__print(*args, logLevel=DEBUG, **kwargs)

    def info(self, *args, **kwargs):
        self.__print(*args, logLevel=INFO, **kwargs)

    def warn(self, *args, **kwargs):
        self.__print(*args, logLevel=WARNING, **kwargs)

    def error(self, *args, **kwargs):
        self.__print(*args, logLevel=ERROR, **kwargs)

    def message(self, *args, **kwargs):
        self.__print(*args, **kwargs)
    
    def logPrefix(self):
        return self.name

    def __print(self, *args, **kwargs):
        args = list(args)
        if 'failure' in kwargs:
            failure = kwargs['failure']
            del kwargs['failure']
            out = StringIO()
            try:
                failure.raiseException()
            except Exception, e:
                out.write("%s\n" % str(e))
            failure.printTraceback(file=out)
            args.append(out.getvalue())
            out.close()

        if 'exc_info' in kwargs:
            exc_info = kwargs['exc_info']
            del kwargs['exc_info']
            args.append(traceback.format_exc())

        if 'logLevel' in kwargs:
            level = kwargs['logLevel']
            del kwargs['logLevel']
        else:
            level = INFO

        if 'period' in kwargs:
            period = kwargs['period']
            del kwargs['period']
        else:
            period = None

        if 'log_msg_id' in kwargs:
            extra = dict(log_msg_id=kwargs['log_msg_id'])
            del kwargs['log_msg_id']
        else:
            extra = None

        if self.log is None:
            self.log = logging.getLogger(self.name)

        kv = []
        for k in kwargs:
            kv.append("%s=%s" % (k, kwargs[k]))

        msg = " ".join((" ".join(args), ",".join(kv)))
        if self.__periodic_check(msg, period):
            self.log.log(level, msg, extra=extra)

    def __periodic_check(self, msg, period):
        # Prune old messages if needed
        time = self.__get_time()
        if self.last_prune_time + REPEATABLE_MESSAGES_PRUNE_TIME <= time:
            self.last_prune_time = time
            for msg_id, next_time in self.repeatable_messages.items():
                if next_time <= time:
                    del self.repeatable_messages[msg_id]

        if period is None:
            return True

        msg_id = hash(msg)
        if msg_id in self.repeatable_messages and \
           self.repeatable_messages[msg_id] > time:
            return False

        self.repeatable_messages[msg_id] = time + period
        return True

    def __get_time(self):
        import time
        return time.time()
        
