#!/usr/bin/env xpyv

#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import sys
import logging.handlers
import platform

from os.path import normpath, join

from hak.common.definitions import PROJECT_ROOT_PATH

def get_libpath():
    return join(normpath(PROJECT_ROOT_PATH), 'lib')

def get_mpath_management_path():
    return join(normpath(PROJECT_ROOT_PATH), 'lib', 'mpath_management')

def set_pythonpath():
    sys.path.append(get_libpath())
    sys.path.append(get_mpath_management_path())

set_pythonpath()

from hak_env_config import initialize_project
initialize_project()

from win32service import SERVICE_STOP_PENDING
from win32serviceutil import HandleCommandLine, ServiceFramework
from win32event import CreateEvent, WaitForSingleObject, WAIT_TIMEOUT

from twisted.internet import reactor, defer

from common.agent import log_msg_id
from common.agent.config import config
from common.agent.manager import Manager
from common.agent.manager import WaitingForSnapshotError
from common.agent.manager import DeviceNotFoundError
from common.agent.manager import UnsupportedOperationError
from common.agent.model import Device, XIVDevice
from common.agent.web_server import run_server
from common.agent.web_resource import AgentWebResource

# This causes an import error.
# Instead of refactoring the entire service runtime away to a happy place, we only mask it, so that we still can use
# the remove, stop and exists part of the service
#TODO: Refactor away all the legacy service code (https://jira.xiv.ibm.com/browse/HA-228211)

#from windows_native_mpio_wrapper import WindowsNativeMPIOWrapper
from windows_config_registry import WindowsConfigRegistry
from set_process_priority import set_process_priority

config.set(WindowsConfigRegistry())

from utils.log import Logger
event_log = Logger("Events")
log = Logger("Service")

def is_win2003():
    return platform.release() == '2003Server'

class CustomizedNTEventLogHandler(logging.handlers.NTEventLogHandler):
    def __init__(self):
        event_log_dll = normpath(join(PROJECT_ROOT_PATH,
            "deps", "blobs", "windows", self.get_architecture(),
            "eventlog", "eventlog.dll"))
        logging.handlers.NTEventLogHandler.__init__(self,
                                                    appname=config.SERVICE_NAME,
                                                    dllname=event_log_dll,
                                                    logtype='System')

    def get_architecture(self):
        arch_str = platform.architecture()[0]
        if arch_str.find("32") != -1:
            return 'x86'
        else:
            return 'x64'

    def getMessageID(self, record):
        if 'log_msg_id' in record.__dict__:
            return record.log_msg_id
        else:
            return 255

class WindowsManager(Manager):
    def __init__(self, *args, **kwargs):
        Manager.__init__(self, *args, **kwargs)
        self.is_win2003 = is_win2003()

    def on_new_snapshot(self, cur_snapshot):
        self.enforce_round_robin_policy_if_needed(cur_snapshot)
        return Manager.on_new_snapshot(self, cur_snapshot)

    def enforce_round_robin_policy_if_needed(self, snapshot):
        if not self.is_win2003:
            return

        for device in snapshot.devices.values():
            if isinstance(device, XIVDevice) \
                   and device.lb_policy != 'RoundRobin':
                self._set_device_to_roundrobin_policy(device)

    def set_lb_policy(self, scsi_id, policy, path_params):
        # Allow only RoundRobin policy for XIV devices in Win2003
        if self.is_win2003 and policy != 'RoundRobin':
            device = self.find_device_from_snapshot_by_scsi_id(scsi_id)
            if isinstance(device, XIVDevice):
                raise UnsupportedOperationError(scsi_id)

        return Manager.set_lb_policy(self, scsi_id, policy, path_params)

    def _set_device_to_roundrobin_policy(self, device):
        event_log.info("Enforcing RoundRobin policy to XIV device:\n%s" %
            (device.summary(),),
            log_msg_id=log_msg_id.ENFORCE_ROUND_ROBIN_POLICY)
        
        d = defer.maybeDeferred(self.native_mpio.set_load_balancing_policy,
                                device.device_file_path, 'RoundRobin', dict())
        def _cb(_):
            device.lb_policy = 'RoundRobin'
        def _eb(failure):
            event_log.warn("Failed to enforce RoundRobin policy to XIV device:\n%s" %
                (device.summary(),), failure=failure,
                log_msg_id=log_msg_id.FAILED_ENFORCE_ROUND_ROBIN_POLICY)
        d.addCallback(_cb)
        d.addErrback(_eb)

    def _on_device_up(self, device):
        if isinstance(device, XIVDevice):
            if self.is_win2003:
                if device.lb_policy != 'RoundRobin':
                    self._set_device_to_roundrobin_policy(device)
            else:
                if device.lb_policy == 'FailOverOnly' \
                   and device.scsi_product_revision.startswith('10.0'):
                    self._set_device_to_roundrobin_policy(device)

        Manager._on_device_up(self, device)

    def _on_device_existing(self, device):
        if isinstance(device, XIVDevice):
            if device.lb_policy == 'FailOverOnly':
                log.warn("Setting fail-over-only load balancing strategy to an XIV device is not recommended " +
                         "since it may have severe performance impact. Please consider using other load-balancing " +
                         "policies that will dramatically improve the performance of the device.\n" +
                         "Device information:\n%s", device.summary(), period=7*60*60*24,
                         log_msg_id=log_msg_id.FAIL_OVER_ONLY_CHOSEN_FOR_XIV_DEVICE)

class WindowsAgentWebResource(AgentWebResource):
    def _get_available_lb_policies(self):
        return Device.LoadBalancingPolicy.keys()

    def _get_machine_info(self):
        from xsf.host.os.windows.info import Windows
        info = Windows()
        uname = AgentWebResource._get_machine_info(self)
        uname['system'] = info.version
        if info.service_pack and info.service_pack > 0:
            uname['system'] += " SP%d" % info.service_pack
        uname['release'] = info.edition
        return uname

class AgentRunner:
    def __init__(self):
        config.init_logging()
        self.init_event_logging()

        # This is the only place the native IO wrapper is used.
        # As we're not running the actuall agent, it doesn't disturb the configure and verify scripts

        #self.manager = WindowsManager(WindowsNativeMPIOWrapper())

        set_process_priority()
        event_log.info("XIV MPIO Agent started.",
                       log_msg_id=log_msg_id.SERVICE_START)

    def init_event_logging(self):
        # We set two handlers here: one specifically for "Events" where the
        # application will explictly log MPIO-related messages, and the other
        # catches warnings and errors and writes them to the event log too.
        formatter = config.get_logging_formatter()

        h = CustomizedNTEventLogHandler()
        h.setFormatter(formatter)
        h.setLevel(logging.WARNING)

        root_log = logging.getLogger("")
        root_log.addHandler(h)

        h = CustomizedNTEventLogHandler()
        h.setFormatter(formatter)
        h.setLevel(logging.INFO)

        event_log = logging.getLogger("Events")
        event_log.propagate = 0
        event_log.addHandler(h)
        event_log.setLevel(logging.INFO)

    def run(self, installSignalHandlers=True):
        run_server(WindowsAgentWebResource(self.manager),
                   installSignalHandlers=installSignalHandlers)

class AgentService(AgentRunner, ServiceFramework):
    _svc_name_ = config.SERVICE_NAME
    _svc_display_name_ = config.SERVICE_DISPLAY_NAME
    _svc_description_ = config.SERVICE_DESCRIPTION
    
    def __init__(self, args):
        AgentRunner.__init__(self)
        ServiceFramework.__init__(self, args)
        self.wait_stop_event = CreateEvent(None, 0, 0, None)

    def SvcDoRun(self):
        self.check_svc_stop()
        self.run(installSignalHandlers=False)

    def SvcStop(self):
        self.ReportServiceStatus(SERVICE_STOP_PENDING)
        reactor.stop()

    def check_svc_stop(self):
        retval = WaitForSingleObject(self.wait_stop_event, 0)
        if retval == WAIT_TIMEOUT:
            reactor.callLater(0.5, self.check_svc_stop)
        else:
            reactor.stop()

if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == 'standalone':
        AgentRunner().run()
    else:
        HandleCommandLine(AgentService)
