#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program  
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
# 
#-------------------------------------------------------------------------------------------------   
import re
import time
import platform

from xsf.host.storage import multipath
from xsf.host.os.windows.wmi_utils import wmi_escape, extract_wmi_value
from xsf.xiv.identify import is_xiv_device

from xsf.host.scsi import raw as scsy_raw
from xsf.host.scsi import commands as scsy_commands
from xsf.host.scsi.errors import ScsiStatusCheckCondition

from common.agent import model
from common.agent import log_msg_id
from common.agent.config import config
from common.agent.native_mpio import NativeMPIO
from common.agent.native_mpio import SetLBPolicyFailedError
from common.agent.native_mpio import NativeMPIONotReadyError

from utils.log import Logger, PERIOD_ONCE

import win32file
import struct
import time

from wmi import x_wmi
import win32com.client

IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS = 5636096

PHYSICAL_DRIVE_PREFIX = '\\\\.\\PHYSICALDRIVE'

log = Logger("NativeMPIO")

LoadBalancingPolicyIDToName = {
    1: 'FailOverOnly',
    2: 'RoundRobin',
    3: 'RoundRobinWithSubset',
    4: 'DynamicLeastQueueDepth',
    5: 'WeightedPaths',
    6: 'LeastBlocks'
}

VolumeDriveType = dict(Unknown=0, NoRootDirectory=1, RemovableDisk=2,
                       LocalDisk=3, NetworkDrive=4, CompactDisk=5, RAMDisk=6)

UNKNOWN_LOAD_BALANCING_POLICY_ID = 100

CACHE_VALIDITY_TIME = 60 * 60

LoadBalancingPolicyNameToID = dict([ (val, key)
    for key,val in LoadBalancingPolicyIDToName.items() ])

# TODO: move this to SCSY
def scsi_left_aligned(s):
    return s.rstrip("\x00 ")


class WindowsNativeMPIO(NativeMPIO):
    def __init__(self, *args, **kwargs):
        NativeMPIO.__init__(self, *args, **kwargs)
        self.pnp_device_id_cache = {}
        self.pnp_device_id_cache_prune_time = time.time()
        self._cached_wmi_com = None
        self._cached_cimv2_com = None

    @property
    def _wmi_root(self):
        if self._cached_wmi_com is None:
            self._cached_wmi_com = win32com.client.GetObject("winmgmts:root\\wmi")
        return self._cached_wmi_com

    @property
    def _cimv2_root(self):
        if self._cached_cimv2_com is None:
            self._cached_cimv2_com = win32com.client.GetObject("winmgmts:root\\cimv2")
        return self._cached_cimv2_com
    
    def set_load_balancing_policy(self, physical_device_id, policy,
                                  path_params):
        # TODO:
        # 1. make sure policy is in the hash (and not unknown)
        pnp_dev_id = self._get_mpio_pnp_device_id(physical_device_id)
        dev_op = self._get_wmi_device_lb_policy_op(pnp_dev_id)
        if dev_op is None:
            log.debug("set load balancing policy: device %s (%s) not found" %
                      (physical_device_id, pnp_dev_id))
            raise SetLBPolicyFailedError(physical_device_id, policy,
                                         "device not found")
        
        log.info("setting load balancing policy of %s (%s) to %s" %
                 (physical_device_id, pnp_dev_id, policy))

        dev_policy = self._get_wmi_lb_device_policy(pnp_dev_id).Properties_.Item('LoadBalancePolicy').Value
        try:
            result = dict()
            dev_policy.Properties_.Item('LoadBalancePolicy').Value = \
                                            LoadBalancingPolicyNameToID[policy]
            paths = dev_policy.Properties_('DSM_Paths').Value
            if policy == 'RoundRobin' or policy == 'DynamicLeastQueueDepth' \
                   or policy == 'LeastBlocks':
                for path in paths:
                    path.Properties_.Item('PrimaryPath').Value = 1
                    path.Properties_.Item('OptimizedPath').Value = 1
                    result[path.Properties_.Item('DsmPathId').Value] = \
                                        dict(primary=True)
            elif policy == 'FailOverOnly':
                if len(path_params) != 1:
                    raise SetLBPolicyFailedError(physical_device_id, policy,
                                                 "must have one path as active")
                active_path_id = path_params.keys()[0]
                found_active_path = False
                for path in paths:
                    if path.Properties_('DsmPathId').Value == active_path_id:
                        result[path.Properties_.Item('DsmPathId').Value] = \
                                                        dict(primary=True)
                        found_active_path = True
                        path.Properties_.Item('PrimaryPath').Value = 1
                        path.Properties_.Item('OptimizedPath').Value = 1
                    else:
                        result[path.Properties_.Item('DsmPathId').Value] = \
                                                        dict(primary=False)
                        path.Properties_.Item('PrimaryPath').Value = 0
                        path.Properties_.Item('OptimizedPath').Value = 0

                if not found_active_path:
                    raise SetLBPolicyFailedError(physical_device_id, policy,
                                "could not find path %s" % active_path_id)
            elif policy == 'WeightedPaths':
                if len(path_params) != len(paths):
                    raise SetLBPolicyFailedError(physical_device_id, policy,
                                        "policy requires all paths have value")
                for path in paths:
                    path_id = path.Properties_('DsmPathId').Value
                    if path_id not in path_params:
                        raise SetLBPolicyFailedError(physical_device_id, policy,
                                "could not find value for path %s" % path_id)
                    weight = int(path_params[path_id])
                    result[path_id] = dict(primary=True, weight=weight)
                    path.Properties_.Item('PrimaryPath').Value = 1
                    path.Properties_.Item('OptimizedPath').Value = 1
                    path.Properties_.Item('PathWeight').Value = weight
            elif policy == 'RoundRobinWithSubset':
                for path in paths:
                    path_id = path.Properties_('DsmPathId').Value
                    is_active = 1
                    if path_id not in path_params or \
                           int(path_params[path_id]) == 0:
                        is_active = 0
                    path.Properties_.Item('PrimaryPath').Value = is_active
                    path.Properties_.Item('OptimizedPath').Value = is_active
                    result[path_id] = dict(primary=(is_active == 1))
                
            params = dev_op.Methods_('DsmSetLoadBalancePolicy').InParameters.SpawnInstance_()
            params.Properties_.Item('LoadBalancePolicy').Value = dev_policy
            dev_op.ExecMethod_('DsmSetLoadBalancePolicy', params)
            return result
        except SetLBPolicyFailedError, e:
            log.warn("failed to set load balancing policy of %s to %s (path params=%s)" %
                     (physical_device_id, policy, repr(path_params)), exc_info=e)
            raise
        except Exception, e:
            log.warn("failed to set load balancing policy of %s (%s) to %s" %
                     (physical_device_id, pnp_dev_id, policy), exc_info=e,
                     log_msg_id=log_msg_id.FAILED_SET_LOAD_BALANCE_POLICY)
            raise SetLBPolicyFailedError(physical_device_id, policy, str(e))

    def query_snapshot(self):
        log.debug("child query_snapshot")
        snapshot = model.Snapshot(timestamp=time.time())
        disk_mount_point_mapping = self._query_disk_mount_point_mapping()

        disks = multipath.get_all_multipath_devices()

        # Give an informative log if there are no MPIO disks.
        if len(disks) == 0:
            devices = "\n".join([ "%s (%s)" % (disk.name, disk.device.win32_mpio_identifier) for disk in disks ])
            log.warn("No Multipath I/O devices were found out of the following devices:\n%s\n(message repeating every 1 hour)" % devices, period=3600,
                     log_msg_id=log_msg_id.NO_MPIO_DEVICES_FOUND)
        
        for device in disks:
            if device.scsi_serial is not None and len(device.scsi_serial.strip()) > 0:
                dev = self._create_model_device(device,
                                                disk_mount_point_mapping)
                snapshot.devices[dev.scsi_address] = dev
            else:
                log.warn("Drive %s has a non-unique SCSI serial number and will not work correctly with MPIO. This drive will be ignored." % device.name,
                         period=PERIOD_ONCE,
                         log_msg_id=log_msg_id.NON_UNIQUE_SCSI_SERIAL)
        return snapshot
        
    def _create_model_device(self, xsf_dev, disk_mount_point_mapping):
        if xsf_dev.is_xiv_device:
            dev = model.XIVDevice()

            from xsf.xiv.xiv_device import XivDevice
            xiv_dev = XivDevice(xsf_dev.device,
                                xsf_dev.scsi_page80,
                                xsf_dev.scsi_page83)

            dev.xiv_machine_serial = xiv_dev.machine_serial_number_decimal
            dev.xiv_vol_serial = xiv_dev.volume_serial_number_decimal
            dev.xiv_vol_name = xiv_dev.volume_name
            dev.xiv_host_name = xiv_dev.host
        else:
            dev = model.Device()

        dev.device_file_path = xsf_dev.name
        dev.scsi_address = xsf_dev.scsi_serial
        dev.scsi_vendor_id = xsf_dev.scsi_standard_inquiry["vendor_identification"]
        # The SCSI spec says these are left-aligned fields. An x-aligned field
        # may contain spaces as padding and may end with one or more NULLs.
        dev.scsi_product_id = scsi_left_aligned(xsf_dev.scsi_standard_inquiry["product_identification"])
        dev.scsi_product_revision = scsi_left_aligned(xsf_dev.scsi_standard_inquiry["product_revision_level"])

        pnp_id = self._get_mpio_pnp_device_id(dev.device_file_path)
        dev.os_info['pnp_device_id'] = pnp_id
        (dev.lb_policy, path_lb_info) = self._get_lb_device_policy(pnp_id)
        dev.supported_lb_policies = self._get_supported_lb_policies(dev, pnp_id)

        drive_num = self._physical_drive_id_to_drive_num(dev.device_file_path)
        if drive_num in disk_mount_point_mapping:
            dev.mount_points = list(disk_mount_point_mapping[drive_num])

        self._create_model_paths(dev, xsf_dev, path_lb_info)
        
        return dev

    def _create_model_paths(self, dev, xsf_dev, path_lb_info):
        paths = {}

        active_paths = filter(lambda d: d.is_path_active,
                              xsf_dev.single_path_devices)

        for child in active_paths:
            if isinstance(dev, model.XIVDevice):
                path = model.XIVDevicePath(id=child.device.win32_mpio_path_id)
                if path.id in path_lb_info:
                    self._set_xiv_path_attributes_from_path_info(path,
                                                        path_lb_info[path.id])
            else:
                path = model.DevicePath(id=child.device.win32_mpio_path_id)
            path.scsi_target = child.target_id
            path.scsi_lun = child.lun
            # FIXME: we need to know the protocol. Unfortunately, the
            # protocol_identifier in the scsi_page83 doesn't give the right
            # result: it gives fibre_channel for iSCSI.
            path.interface_type = path.interface_type or 'iSCSI'
            if path.id in path_lb_info:
                path.primary = path_lb_info[path.id]['primary']
                path.weight = path_lb_info[path.id]['weight']
            paths[path.id] = path

        dev.device_paths = paths
        self._set_device_path_performance_info(dev)

    def _get_mpio_pnp_device_id(self, physical_device_id):
        # This function is VERY slow for some unknown reason. That's why we'll
        # be caching the results.
        now = time.time()
        if (self.pnp_device_id_cache_prune_time + CACHE_VALIDITY_TIME) < now:
            self.pnp_device_id_cache = {}
            self.pnp_device_id_cache_prune_time = now
        
        if (physical_device_id in self.pnp_device_id_cache):
            return self.pnp_device_id_cache[physical_device_id]
        
        from xsf.host.storage.multipath.windows import get_mpio_device_of_physical_device
        result = get_mpio_device_of_physical_device(physical_device_id)

        self.pnp_device_id_cache[physical_device_id] = result
        return result

    def _get_lb_device_policy(self, pnp_dev_id):
        try:
            obj = self._get_wmi_lb_device_policy(pnp_dev_id)
            if obj is None:
                log.warn("couldn't find device load balancing policy " \
                         "for PnP ID %s" % pnp_dev_id,
                         log_msg_id=log_msg_id.FAILED_FIND_DEVICE_LOAD_BALANCE_POLICY)
                return (None, dict())

            lb_id = obj.LoadBalancePolicy.LoadBalancePolicy
            if lb_id == UNKNOWN_LOAD_BALANCING_POLICY_ID:
                return (None, dict())
        
            if lb_id in LoadBalancingPolicyIDToName:
                path_lb_info = dict()
                for path in obj.LoadBalancePolicy.Properties_('DSM_Paths').Value:
                    path_id = path.Properties_.Item('DsmPathId').Value
                    info = dict()
                    path_lb_info[path_id] = info
                    info['primary'] = int(path.Properties_.Item('PrimaryPath').Value) == 1
                    info['weight'] = int(path.Properties_.Item('PathWeight').Value)
                    try:
                        i = path.Properties_.Item('TargetPort_Identifier').Value
                        iscsi_port = (i & 0xFF) >= 0x20
                        info['machine_port'] = i & 0x1F
                        info['machine_module'] = (i >> 8) & 0x1F
                        info['machine_rack'] = (i >> 13) + 1
                        info['interface_type'] = 'iSCSI' if iscsi_port else 'FiberChannel'
                    except:
                        pass
                    
                return (LoadBalancingPolicyIDToName[lb_id], path_lb_info)
            else:
                log.warn("unknown load balancing policy ID " \
                         "discovered: %s" % lb_id,
                         log_msg_id=log_msg_id.UNKNOWN_LOAD_BALANCE_POLICY)
                return (None, dict())
        except Exception, e:
            log.warn("failed getting load balancing policy " \
                     "for PnP ID %s (message repeating every 1 hour)" %
                     pnp_dev_id, period=3600, exc_info=e,
                     log_msg_id=log_msg_id.FAILED_GET_LOAD_BALANCE_POLICY)
            return (None, dict())

    def _set_xiv_path_attributes_from_path_info(self, path, path_info):
        for prop in [ 'machine_module', 'machine_port', 'machine_rack',
                      'interface_type' ]:
            if prop in path_info:
                setattr(path, prop, path_info[prop])

    def _query_disk_mount_point_mapping(self):
        try:
            drives_to_mount_points = {}
            for obj in self._cimv2_root.ExecQuery("SELECT * FROM Win32_MountPoint"):
                try:
                    device_id = extract_wmi_value("Win32_Volume.DeviceID", obj.Volume)
                    mount_point = extract_wmi_value("Win32_Directory.Name", obj.Directory)
                except Exception, e:
                    continue

                res = self._cimv2_root.ExecQuery("SELECT * FROM Win32_Volume WHERE DeviceID=\"%s\"" % wmi_escape(device_id))
                if len(res) != 1:
                    continue

                vol = res[0]

                if vol.DriveType != VolumeDriveType['LocalDisk']:
                    # MPIO devices are marked as local disks
                    continue

                try:
                    drives = self._get_volume_drive_extents(device_id)
                    for drive in drives:
                        if drive not in drives_to_mount_points:
                            drives_to_mount_points[drive] = set()
                        drives_to_mount_points[drive].add(mount_point)
                except:
                    log.info("failed to fetch volume drive extents information for volume %s" % device_id)
            return drives_to_mount_points
        except Exception, e:
            log.warn("failed to fetch mount point information", exc_info=e,
                     log_msg_id=log_msg_id.FAILED_FETCH_MOUNT_POINT)
            return {}
            
    def _get_volume_drive_extents(self, volume_device_id):
        handle = None
        volume_device_id = volume_device_id.replace("?", ".")
        if volume_device_id[-1] == "\\":
            volume_device_id = volume_device_id[0:-1]
        try:
            # FIXME: find all volumes first using WMI (CIMV2 Win32_Volume).
            handle = win32file.CreateFile(volume_device_id, 0, 0, None,
                                          win32file.OPEN_EXISTING, 0, 0)
            raw_data = win32file.DeviceIoControl(handle,
                            IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS, None, 4096)
            (num_extents,) = struct.unpack('=Lxxxx', raw_data[0:8])
            offset = 8
            elem_size = struct.calcsize('=LQQ')

            volume_drives = []
            for i in xrange(0, num_extents):
                (drive_num, start_ofs, extent_len) = \
                            struct.unpack_from('=LQQ', raw_data, offset=offset)
                offset += elem_size
                volume_drives.append(drive_num)

            return volume_drives
        finally:
            win32file.CloseHandle(handle)

    def _get_wmi_lb_device_policy(self, pnp_dev_id):
        query_obj = 'DSM_QueryLBPolicy_V2'
        try:
            result = self._wmi_root.Get(query_obj)
        except:
            query_obj = 'DSM_QueryLBPolicy'

        result_set = self._wmi_root.ExecQuery("SELECT * FROM %s WHERE InstanceName=\"%s\"" % (query_obj, wmi_escape(pnp_dev_id)))
        if len(result_set) == 0:
            return None
        return result_set[0]

    def _get_supported_lb_policies(self, dev, pnp_dev_id):
        if isinstance(dev, model.XIVDevice) and platform.release() == '2003Server':
             return [ 'RoundRobin' ]
        return self._get_wmi_supported_lb_policies(pnp_dev_id)
        
    def _get_wmi_supported_lb_policies(self, pnp_dev_id):
        result = self._wmi_root.ExecQuery("SELECT * FROM DSM_QuerySupportedLBPolicies WHERE InstanceName=\"%s\"" % wmi_escape(pnp_dev_id))
        if len(result) == 0:
            return []
        policies = []
        for policy in result[0].Properties_.Item('Supported_LB_Policies').Value:
            policy = int(policy.LoadBalancePolicy)
            if policy in LoadBalancingPolicyIDToName:
                policies.append(LoadBalancingPolicyIDToName[policy])
        return policies
        
    def _get_wmi_device_lb_policy_op(self, pnp_dev_id):
        return self.__find_by_instance_name(pnp_dev_id,
                                        self._get_wmi_device_lb_policy_ops())

    def _get_wmi_device_lb_policy_ops(self):
        return self._wmi_root.ExecQuery("SELECT * FROM DSM_LB_Operations")

    def _physical_drive_id_to_drive_num(self, drive_id):
        if not drive_id.startswith(PHYSICAL_DRIVE_PREFIX):
            return None
        return int(drive_id[len(PHYSICAL_DRIVE_PREFIX):])

    def _set_device_path_performance_info(self, device):
        # Try two methods: one is MPIO_PATH_HEALTH and the other one is
        # MSDSM_DEVICE_PERF. We prefer MPIO_PATH_HEALTH but it's only for
        # 2008R2 and there too it can be disabled. MSDSM_DEVICE_PERF is for 2008
        # - there are no counters for 2003.
        if platform.release() != '2003Server':
            try:
                self._set_device_path_performance_info_mpio_path_health(device)
            except Exception, e:
                try:
                    self._set_device_path_performance_info_msdsm_device_perf(device)
                except Exception, e:
                    pass
    
    def _set_device_path_performance_info_mpio_path_health(self, device):
        pnp_id = device.os_info['pnp_device_id']
        rs = self._wmi_root.ExecQuery("SELECT * FROM MPIO_DEVINSTANCE_HEALTH_INFO WHERE InstanceName='%s'" % wmi_escape(pnp_id))
        latest_packet_per_path = dict()
        for packet in rs[0].DevInstanceHealthPackets:
            # Take only the latest packet (if I understand how this mechanism
            # works):
            if packet.PathId not in latest_packet_per_path \
               or latest_packet_per_path[packet.PathId] < packet.CreateTime:
                latest_packet_per_path[packet.Pathid] = packet.CreateTime
                if packet.PathId in device.device_paths:
                    path = device.device_paths[packet.PathId]
                    path.io_bytes_read = int(packet.NumberBytesRead)
                    path.io_bytes_written = int(packet.NumberBytesWritten)
                    path.io_read_ops = int(packet.NumberReads)
                    path.io_write_ops = int(packet.NumberWrites)
                    path.io_errors = int(packet.NumberIoErrors)
                    path.io_retries = int(packet.NumberRetries)

    def _set_device_path_performance_info_msdsm_device_perf(self, device):
        pnp_id = device.os_info['pnp_device_id']
        rs = self._wmi_root.ExecQuery("SELECT * FROM MSDSM_DEVICE_PERF WHERE InstanceName='%s'" % wmi_escape(pnp_id))
        for perf_info in rs[0].PerfInfo:
            if perf_info.PathId in device.device_paths:
                path = device.device_paths[perf_info.PathId]
                path.io_bytes_read = int(perf_info.BytesRead)
                path.io_bytes_written = int(perf_info.BytesWritten)
                path.io_read_ops = int(perf_info.NumberReads)
                path.io_write_ops = int(perf_info.NumberWrites)
        
    def __find_by_instance_name(self, name, sequence):
        for obj in sequence:
            # Case-insensitive comparison is a must here!
            if obj.Properties_.Item('InstanceName').Value.upper() == name.upper():
                return obj
        return None




