#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program  
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
# 
#-------------------------------------------------------------------------------------------------   
import os
import sys
import random
import time

if __name__ == '__main__':
    from hak.common.definitions import PROJECT_ROOT_PATH
    from os.path import join, normpath

    def get_libpath():
        return join(normpath(PROJECT_ROOT_PATH), 'lib')

    def get_mpath_management_path():
        return join(normpath(PROJECT_ROOT_PATH), 'lib', 'mpath_management')

    def set_pythonpath():
        sys.path.append(get_libpath())
        sys.path.append(get_mpath_management_path())

    set_pythonpath()

    from hak_env_config import initialize_project
    initialize_project()


    # Ignore all warnings of type DeprecationWarning since Twisted issues them
    import warnings
    warnings.filterwarnings('ignore', category=DeprecationWarning)

from twisted.internet import reactor, protocol, error, stdio
from twisted.spread import pb

from common.agent import model
from common.agent import log_msg_id
from common.agent.xml_serialization import XMLSerializer, XMLDeserializer
from common.agent.config import config
from common.agent.native_mpio import NativeMPIO, NativeMPIONotReadyError

from windows_native_mpio import WindowsNativeMPIO
from set_process_priority import set_process_priority

from utils.log import Logger

CHILD_LIFETIME = 0 # 10 * 60

if __name__ == '__main__':
    log = Logger("NativeMPIOChild")
else:
    log = Logger("NativeMPIOWrapper")

class ProcessFailedError(Exception):
    def __init__(self, status, data):
        Exception.__init__(self, "process failed with status code: %s and stdout data: %s" % (status, data))
        self.status = status
        self.data = data

HEX_DIGITS = "0123456789abcdef"
def mk_shared_secret():
    t = int(time.time())
    s = []
    for i in xrange(0, 20):
        s.append(HEX_DIGITS[(((t >> (i * 4)) & 0xf) + random.randint(0, 15)) % 16])
    return ''.join(s)

class ParentProcessProtocol(protocol.ProcessProtocol):
    def __init__(self, wrapper):
        self.wrapper = wrapper

    def connectionMade(self):
        self.transport.write("%d\n%s\n" % (self.wrapper.port_number, self.wrapper.shared_secret))

    def outReceived(self, data):
        pass

    def errReceived(self, data):
        log.warn("unexpected stderr data from client: %s" % data,
                 log_msg_id=log_msg_id.UNEXPECTED_STDERR_DATA_FROM_CLIENT)

    def processEnded(self, status):
        if status.check(error.ProcessDone):
            self.wrapper.process_ended(status)
        else:
            self.wrapper.process_died(status)

from twisted.spread import jelly
class AllowEverything(jelly.SecurityOptions):
    def isModuleAllowed(self, modulename):
        return True
    def isClassAllowed(self, klass):
        return True
    def isTypeAllowed(self, typeName):
        return True

class WindowsNativeMPIOServer(pb.Referenceable, WindowsNativeMPIO):
    def __init__(self):
        WindowsNativeMPIO.__init__(self)

    remote_query_snapshot = WindowsNativeMPIO.query_snapshot
    remote_set_load_balancing_policy = WindowsNativeMPIO.set_load_balancing_policy

class WindowsNativeMPIOWrapper(NativeMPIO, pb.Root):
    def __init__(self):
        self.xpyv_path = self._get_xpyv_exe_path()
        self.this_file_path = os.path.realpath(__file__)
        self.factory = pb.PBClientFactory(security=AllowEverything())
        self.shared_secret = mk_shared_secret()
        self.remote = None

        port_obj = reactor.listenTCP(port=0, interface='localhost',
                factory=pb.PBServerFactory(self, security=AllowEverything(), 
                                           unsafeTracebacks=True))
        self.port_number = port_obj.socket.getsockname()[1]

        self._spawn_child()

    def query_snapshot(self):
        self.validate_child_alive()

        d = self.remote.callRemote("query_snapshot")
        d.addErrback(self.connectivity_error)
        return d

    def set_load_balancing_policy(self, device_id, policy, policy_params):
        self.validate_child_alive()

        d = self.remote.callRemote("set_load_balancing_policy",
                                   device_id, policy, policy_params)
        d.addErrback(self.connectivity_error)
        return d

    def remote_client_connected(self, remote, shared_secret):
        log.info("child connected")
        if self.shared_secret == shared_secret:
            def reset_remote(remote_ref):
                self.remote = None
            self.remote = remote
            self.remote.notifyOnDisconnect(reset_remote)
        else:
            log.info("client tried to connect with an incorrect shared secret (server %s != client %s)",
                     self.shared_secret, shared_secret)

    def validate_child_alive(self):
        if self.remote is None:
            raise NativeMPIONotReadyError()

    def process_died(self, status):
        log.error("child process died, starting another one in 10 seconds",
                  status=status, log_msg_id=log_msg_id.CHILD_PROCESS_DIED)
        self.remote = None
        reactor.callLater(10, self._spawn_child)

    def process_ended(self, status):
        log.info("child process finished w/o errors, spawning another one")
        # FIXME: check if shutting down
        self.remote = None
        self._spawn_child()
    
    def _spawn_child(self):
        protocol = ParentProcessProtocol(self)
        env = os.environ.copy()
        sys_path = list(sys.path)
        sys_path.append(os.path.normpath(os.path.join(os.path.realpath(__file__), "..", "..", "..", "..", "..")))
        env['PYTHONPATH'] = os.pathsep.join(sys_path)
        args = [ self.xpyv_path, self.this_file_path, "-c", "connect" ]
        log.debug("spawning process: %s" % " ".join(args), sys_path=sys_path)
        return reactor.spawnProcess(protocol, self.xpyv_path, args, env)

    # TODO: merge with XIVmpioService (manager)
    def _get_xpyv_exe_path(self):
        return self._search_for_file_from_environment('xpyv.exe', 'PATH')

    # TODO: merge with XIVmpioService (manager)
    def _search_for_file_from_environment(self, filename, environment_variable):
        for item in os.environ[environment_variable].split(';'):
            filepath = os.path.join(item, filename)
            if os.path.isfile(os.path.join(filepath)):
                return filepath
        return None 

    def connectivity_error(self, failure):
        if failure.check(pb.PBConnectionLost, error.ConnectionDone):
            # Verify that it holds a ConnectionDone inside (and then we ignore
            # it because it's part of a normal child recycling).
            try:
                failure.raiseException()
            except pb.PBConnectionLost, e:
                failure = e.args[0]
            except error.ConnectionDone:
                raise NativeMPIONotReadyError()
        return failure

class StopOnStdinCloseWatchdog(protocol.Protocol):
    def connectionLost(self, *args):
        log.debug("Parent closed pipe, exiting.")
        reactor.stop()
        
def main(argv):
    def query_snapshot():
        snapshot = WindowsNativeMPIO().query_snapshot()
        print XMLSerializer().serialize(snapshot)

    def set_lb_policy(dev, lb_policy, policy_params):
        WindowsNativeMPIO().set_load_balancing_policy(dev, lb_policy,
                                                      policy_params)
        print "Load balancing policy %s was set for device %s" % (dev,
                                                                  lb_policy)
        print "Path parameters:\n%s" % policy_params

    def connect():
        from windows_config_registry import WindowsConfigRegistry
        config.set(WindowsConfigRegistry())
        config.init_logging('child_log_file_name')

        port = int(sys.stdin.readline().strip())
        shared_secret = sys.stdin.readline().strip()

        server = WindowsNativeMPIOServer()

        factory = pb.PBClientFactory(security=AllowEverything(), 
                                     unsafeTracebacks=True)
        reactor.connectTCP("localhost", port, factory)
        d = factory.getRootObject()

        def _cb(obj):
            obj.callRemote("client_connected", server, shared_secret)
            # Patch to handle leaky WMI module:
            def stop_reactor():
                if len(obj.broker.waitingForAnswers) > 0:
                    reactor.callLater(0.1, stop_reactor)
                    return
                log.info("child ending so it can restart.")
                reactor.stop()
            if CHILD_LIFETIME > 0:
                reactor.callLater(CHILD_LIFETIME, stop_reactor)

        def _eb(failure):
            log.error("Failed to connect to parent - exiting.",
                      failure=failure)
            os._exit(1)

        d.addCallback(_cb)
        d.addErrback(_eb)

        stdio.StandardIO(StopOnStdinCloseWatchdog())

        set_process_priority()

        reactor.run()

    from optparse import OptionParser, OptionGroup
    usage = "%prog [options]"
    # FIXME: print the commands and their options as well.
    parser = OptionParser(usage=usage)
    commands = OptionGroup(parser,
        "Available commands",
        "The following commands are available: query_snapshot, set_lb_policy, serve")
    commands.add_option("-c", "--command", dest="command",
                        choices=('query_snapshot', 'set_lb_policy', 'connect'))
    parser.add_option_group(commands)

    set_lb_policy_opts = OptionGroup(parser,
        "Command \"set_lb_policy\" (Set Load Balancing Policy)",
        "The following options are available for the set_lb_policy command")
    set_lb_policy_opts.add_option("-d", "--device", dest="dev")
    set_lb_policy_opts.add_option("-p", "--policy", dest="lb_policy",
                            choices=model.Device.LoadBalancingPolicy.keys())
    parser.add_option_group(set_lb_policy_opts)
    (options, args) = parser.parse_args()

    try:
        if options.command == 'query_snapshot':
            query_snapshot()
        elif options.command == 'set_lb_policy':
            lb_policy_params = dict(map(lambda e: e.split("="), args))
            set_lb_policy(options.dev, options.lb_policy, lb_policy_params)
        elif options.command == 'connect':
            connect()
    except Exception, e:
        import traceback
        traceback.print_exc()
        sys.exit(1)

if __name__ == '__main__':
    main(sys.argv)
